import sys
from pathlib import Path
from locale import getpreferredencoding


class Spacy2ConllParser:
    def __init__(self, model='en_core_web_sm', nlp=None, verbose=False):
        self.h_out = None
        if nlp is not None:
            self.nlp = nlp
        else:
            model = __import__(model)
            self.nlp = model.load()

        # To get the morphological info, we need a tag map
        self.tagmap = self.nlp.Defaults.tag_map

        self.verbose = verbose
        self.include_headers = False

    def _close_h_out(self):
        # Close output stream
        if self.h_out is not sys.stdout:
            self.h_out.close()

    def _get_morphology(self, tag):
        if not self.tagmap or tag not in self.tagmap:
            return '_'
        else:
            feats = [f'{prop}={val}' for prop, val in self.tagmap[tag].items() if not Spacy2ConllParser._is_number(prop)]
            if feats:
                return '|'.join(feats)
            else:
                return '_'

    def _iterate(self, text):
        line_idx = 0
        for line in text:
            doc = self.nlp(line.strip())
            last_idx = line_idx
            for idx, parsed_sent in self._sentences_to_conllu(doc, line_idx):
                yield parsed_sent
                last_idx = idx
            line_idx = last_idx

    def _open_h_out(self, output_file, output_encoding):
        # Open output stream
        if output_file:
            self.h_out = open(Path(output_file).resolve(), mode='w', encoding=output_encoding)
        else:
            self.h_out = sys.stdout

    def _sentences_to_conllu(self, doc, line_idx):
        for sent in doc.sents:
            line_idx += 1
            parsed_sent = ''

            if self.include_headers:
                parsed_sent = f'# sent_id = {str(line_idx)}\n'
                parsed_sent += f'# text = {sent.sent}\n'

            for idx, word in enumerate(sent, 1):
                if word.dep_.lower().strip() == 'root':
                    head_idx = 0
                else:
                    head_idx = word.head.i + 1 - sent[0].i

                line_tuple = (
                    idx,
                    word.text,
                    word.lemma_,
                    word.pos_,
                    word.tag_,
                    self._get_morphology(word.tag_),
                    head_idx,
                    word.dep_,
                    '_',
                    '_'
                )
                parsed_sent += '\t'.join(map(lambda x: str(x), line_tuple)) + '\n'

            if self.h_out is not sys.stdout and self.verbose:
                print(parsed_sent)

            yield line_idx, parsed_sent

    def parse(self, input_file=None, input_str=None, input_encoding=getpreferredencoding(), include_headers=False):
        inp_p, inp_str = self._set_input(input_file, input_str)
        self.include_headers = include_headers

        if inp_p:
            with open(inp_p, mode='r', encoding=input_encoding) as fhin:
                yield from self._iterate(fhin)
        else:
            yield from self._iterate(inp_str.split('\n'))

    def parseprint(self, input_file=None, input_str=None, input_encoding=getpreferredencoding(),
                   output_file=None, output_encoding=getpreferredencoding(), include_headers=False):
        self._open_h_out(output_file, output_encoding)

        # Get parsed sentences
        for parsed_sent in self.parse(input_file=input_file, input_str=input_str, input_encoding=input_encoding,
                                      include_headers=include_headers):
            self.h_out.write(parsed_sent + '\n')

        self._close_h_out()

    @staticmethod
    def _set_input(input_file, inp_str):
        if input_file:
            inp_p = Path(input_file).resolve()
            if not inp_p.exists() or not inp_p.is_file():
                raise ValueError(f"'input_file' must be a file. '{str(inp_p)}' given.")
            return inp_p, None
        elif inp_str:
            return None, inp_str
        else:
            raise ValueError("No input given. Use 'input_file' or 'input_str'.")

    @staticmethod
    def _is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False



from setuptools import setup

setup(
    name='spacy_conll',
    version='0.0.3',
    description='Parses a given input text into CoNLL-U format by transforming spaCy output.',
    keywords='nlp spacy conll conllu tagging',
    packages=['spacy_conll'],
    url='https://github.com/BramVanroy/spacy_conll',
    author='Bram Vanroy, Raquel G. Alhama',
    author_email='bramvanroy@hotmail.com, rgalhama@gmail.com',
    license='BSD 2',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Text Processing',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "Operating System :: OS Independent"
    ],
    project_urls={
        'Bug Reports': 'https://github.com/BramVanroy/spacy_conll/issues',
        'Source': 'https://github.com/BramVanroy/spacy_conll',
    },
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['main_parse=spacy_conll.__main__:main']
    }
)



def main():
    import argparse
    from locale import getpreferredencoding
    from spacy_conll import Spacy2ConllParser
    # Input arguments
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description='Parse an input string or input file to CoNLL format.')
    parser.add_argument('-f', '--input_file', default=None,
                        help="Path to file with sentences to parse. Has precedence over 'input_str'.")
    parser.add_argument('-a', '--input_encoding', default=getpreferredencoding(),
                        help="Encoding of the input file. Default value is system default.")
    parser.add_argument('-b', '--input_str', default=None, help="Input string to parse.")
    # Output arguments
    parser.add_argument('-o', '--output_file', default=None,
                        help="Path to output file. If not specified, the output will be printed on standard output.")
    parser.add_argument('-c', '--output_encoding', default=getpreferredencoding(),
                        help="Encoding of the output file. Default value is system default.")
    # Model arguments
    parser.add_argument('-m', '--model', default='en_core_web_sm', help="spaCy model to use.")
    parser.add_argument('-n', '--nlp', default=None,
                        help="Optional already initialised spaCy NLP model. Has precedence over 'model'.")
    # Additional arguments
    parser.add_argument('-v', '--verbose', default=False, action="store_true",
                        help="To print the output to stdout, regardless of 'output_file'.")
    parser.add_argument('-d', '--include_headers', default=False, action="store_true",
                        help="To include headers before every sentence's output. These headers include the sentence"
                             " text and the sentence ID.")

    args = parser.parse_args()

    spacyconll = Spacy2ConllParser(model=args.model, nlp=args.nlp, verbose=args.verbose)

    args = vars(args)
    # Remove 'model' and 'nlp' from the arguments
    for k in ['model', 'nlp', 'verbose']:
        args.pop(k, None)

    spacyconll.parseprint(**args)


if __name__ == '__main__':
    main()
